/* $Id: nit4open.c,v 1.2 89/07/11 03:07:45 dupuy Exp $ */

/*
 * Returns file descriptor for ethernet device by name ("ie0", "le0", etc.).
 * If name is NULL, uses primary ethernet interface.  Will only receive packets
 * of type specified.  Will receive packets for the ethernet address specified,
 * or local ethernet address if NULL.  If there is an error, returns (-1) and
 * the appropriate value is left in errno.  Normal return status zero.
 * Requires superuser privilege.
 */

#include <strings.h>
#include <errno.h>

#include <sys/file.h>
#include <sys/stropts.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>

#include <net/if.h>
#include <net/nit_if.h>
#include <net/nit_pf.h>
#include <net/packetfilt.h>

#include <net/nit.h>			/* NT_ALLTYPES hack */
#include <netinet/in.h>			/* htons() macro */

#include "libether.h"

#ifndef NIT_DEV
#define NIT_DEV "/dev/nit"
#endif

static ether_addr promiscuous;

static short ether_types[FD_SETSIZE];

#define ether_type (ether_types[fd])

#ifndef MULTICAST
static ether_addr bcast_addr = { { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff } };
#endif


static int
i_push (fd)
int fd;
{
    int saved_errno;

    /* push packet filtering module */

    if (ioctl (fd, I_PUSH, "pf") < 0)
    {
	saved_errno = errno;
#ifdef DEBUG
	perror ("ether_open: ioctl I_PUSH");
#endif
	close (fd);
	errno = saved_errno;
	return (-1);
    }

    return (0);
}

static int
i_str (fd, cmd, ptr, len)
int fd;
int cmd;
char *ptr;
int len;
{
    struct strioctl si;
    int saved_errno;

    /* package up the ioctl and issue it */

    si.ic_cmd = cmd;
    si.ic_timout = INFTIM;
    si.ic_len = len;
    si.ic_dp = ptr;
    if (ioctl (fd, I_STR, (char *) &si) < 0)
    {
	saved_errno = errno;
#ifdef DEBUG
	perror ("ether_open: ioctl I_STR");
#endif
	close (fd);
	errno = saved_errno;
	return (-1);
    }

    return (0);
}

static int
niocsetf (fd, filter)
int fd;
struct packetfilt *filter;
{
    struct strioctl si;

    filter->Pf_Priority = 5;		/* unimportant, so long as > 2 */

    return (i_str (fd, NIOCSETF, (char *) filter, sizeof (struct packetfilt)));
}

int
ether_open (name, type, address)
char *name;
int type;
ether_addr *address;
{
    int fd;
    struct strioctl si;
    struct ifreq ifr;
    struct packetfilt filter;
    unsigned short *fptr = 0;
    char **interfaces;
    int saved_errno;
    register int i;

    if (name == 0)			/* get default ethernet interface */
    {
	interfaces = ether_interfaces ();
	if (interfaces == 0 || *interfaces == 0)
	    return (-1);

	name = *interfaces;		/* just use the first name in list */
    }

    if ((fd = open (NIT_DEV, O_RDWR)) < 0)
    {
#ifdef DEBUG
	perror (NIT_DEV);
#endif
	return (-1);
    }

    if (fd >= FD_SETSIZE)		/* not worth making it work here */
    {
        close (fd);
        errno = EMFILE;
	return (-1);
    }

    /* arrange to get discrete messages from the stream */

    if (ioctl (fd, I_SRDOPT, (char *) RMSGD) < 0)
    {
	saved_errno = errno;
#ifdef DEBUG
	perror ("ether_open: ioctl I_SRDOPT");
#endif
	close (fd);
	errno = saved_errno;
	return (-1);
    }

    /* set up filter */

    if (type != NT_ALLTYPES)		/* hack for NIT features */
    {
	if (i_push (fd) < 0)
	    return (-1);

	fptr = &filter.Pf_Filter[0];

	*fptr++ = ENF_PUSHWORD + ETHER_TYPE / sizeof (short);
	*fptr++ = ENF_PUSHLIT | ENF_EQ;
	*fptr++ = htons (type);		/* should be in host order */

	filter.Pf_FilterLen = fptr - &filter.Pf_Filter[0];

	if (niocsetf (fd, &filter) < 0)
	    return (-1);
    }

    /*
     * We defer the bind until after we've pushed the filter to prevent our
     * being flooded with extraneous packets
     */

    strncpy (ifr.ifr_name, name, sizeof (ifr.ifr_name));

    if (i_str (fd, NIOCBIND, (char *) &ifr, sizeof (ifr)) < 0)
	return (-1);

    if (address != 0)
    {
	if (bcmp (address, &promiscuous, sizeof (ether_addr)) != 0)
	{
#ifndef MULTICAST
	    ether_addr local_addr;
#endif
	    if (!ETHER_MCAST (address))
	    {
#ifdef DEBUG
		printf ("rejecting non-multicast address argument\n");
#endif
		close (fd);
		errno = EINVAL;
		return (-1);
	    }

#ifdef MULTICAST
#ifdef __GNUC__
#error not clear how to set multicast addresses
#endif
	    exit (-1);			/* die die die */
#else
	    /*
	     * Enable address matching filter before we go into promiscuous
	     * mode.  We have to do this after the bind, since we can't get the
	     * interface address until then.
	     */

	    (void) ether_address (fd, &local_addr);

	    /* may not have pushed packet filtering module yet */

	    if (fptr == 0)
		if (i_push (fd) < 0)
		    return (-1);

	    fptr = &filter.Pf_Filter[0];

	    if (type != NT_ALLTYPES)	/* hack for NIT features */
	    {
		*fptr++ = ENF_PUSHWORD + ETHER_TYPE / sizeof (short);
		*fptr++ = ENF_PUSHLIT;
		*fptr++ = htons (type);	/* should be in host order */
		*fptr++ = ENF_CAND;
	    }

	    for (i = 0; i < 3; i++)	/* compare addresses in 3 shorts */
	    {
		*fptr++ = ENF_PUSHWORD + ETHER_DST / sizeof (short) + i;
		*fptr++ = ENF_PUSHLIT | ENF_EQ;
		*fptr++ = ((short *) address)[i];
		*fptr++ = ENF_PUSHWORD + ETHER_DST / sizeof (short) + i;
		*fptr++ = ENF_PUSHLIT | ENF_EQ;
		*fptr++ = ((short *) &bcast_addr)[i];
		*fptr++ = ENF_OR;
		*fptr++ = ENF_PUSHWORD + ETHER_DST / sizeof (short) + i;
		*fptr++ = ENF_PUSHLIT | ENF_EQ;
		*fptr++ = ((short *) &local_addr)[i];
		*fptr++ = ENF_CNOR;
	    }

	    filter.Pf_FilterLen = fptr - &filter.Pf_Filter[0];

	    if (niocsetf (fd, &filter) < 0)
		return (-1);

#endif /* MULTICAST */

	}

#ifdef MULTICAST
	else				/* only promiscuous if requested */
#endif

	{				/* go into promiscuous mode */
	    long flag;

	    flag = NI_PROMISC;

	    if (i_str (fd, NIOCSFLAGS, (char *) &flag, sizeof (flag)) < 0)
		return (-1);
	}
    }

    ether_type = type;
    return (fd);
}


/*
 * Writes a single packet, filling src and type fields.
 */

int
ether_write (fd, packet)
int fd;
ether_packet *packet;
{
    struct sockaddr sa;
    struct strbuf cbuf;
    struct strbuf dbuf;

    sa.sa_family = AF_UNSPEC;
    bcopy (packet, sa.sa_data, ETHER_TYPE);

    sa.sa_data[12] = (ether_type >> 8) & 0xff;
    sa.sa_data[13] = ether_type & 0xff;

    cbuf.maxlen = cbuf.len = sizeof (sa);
    cbuf.buf = (char *) &sa;

    /*
     * the interface output routines will paste the ether header back onto the
     * front of the message
     */

    dbuf.maxlen = dbuf.len = packet->pktlen;
    dbuf.buf = packet->pktbuf;

    return (putmsg (fd, &cbuf, &dbuf, 0));
}
